import logging
import os.path

LOGGER = logging.getLogger(__name__)


class Stat(object):

    def __init__(self, username, email, timestamp, collection,
                 product_id, product_date, product_size):

        self.username = username
        self.email = email
        self.timestamp = timestamp
        self.collection = collection
        self.product_id = product_id
        self.product_date = product_date
        self.product_size = product_size

    @staticmethod
    def from_es_record(es_record):
        request = es_record['_source']['request']
        auth = es_record['_source']['auth']
        if '@' in auth:
            username = None
            email = auth
        else:
            username = auth
            email = None
        return Stat(username,
                    email,
                    Stat._str_to_datetime(es_record['_source']['@timestamp']),
                    Stat._get_collection(request),
                    Stat._get_product_id(request),
                    Stat._get_product_date(request),
                    es_record['_source']['bytes'])

    @staticmethod
    def _clean_request(request):
        import re
        # remove download prefix
        request = re.sub('.*download/{1,}', '', request)
        # replace 2 or more slashes by one
        return re.sub('/{2,}', '/', request)

    @staticmethod
    def _str_to_datetime(timestamp):
        import dateutil.parser
        return dateutil.parser.isoparse(timestamp)

    @staticmethod
    def _get_collection(request):
        request = Stat._clean_request(request)
        url_parts = request.split('/')
        if url_parts[0] in ['Sentinel2', 'Sentinel5P']:
            return url_parts[0] + '_' + url_parts[1]
        elif url_parts[0] == "PROBAV":
            if url_parts[1].endswith("V1"):
                return url_parts[0] + '_' + url_parts[1]
            elif url_parts[1] == "C2":
                prod_format = url_parts[2]
                prod_type = url_parts[3].replace('_L3_', '_')
                version = "V2"
                return "_".join([prod_type, prod_format, version])
            else:
                raise Exception(f'Unsupported PROBA-V URL {request}')
        elif url_parts[0] in ['DEM']:
            return url_parts[1]
        elif url_parts[0] == "SPOTVGT":
            return url_parts[0] + '_' + url_parts[1] + '_' + url_parts[2].replace("Free", "")
        elif url_parts[0] == "WORLDCOVER":
            return url_parts[1] + '_' + url_parts[2]
        elif url_parts[0] == "WORLDCEREAL":
            return url_parts[1]
        elif url_parts[0] == "Sentinel3":
            return url_parts[0] + '_' + url_parts[2]
        elif url_parts[1] == "MAPEO_WATER":
            return url_parts[1] + '_' + url_parts[2]
        elif url_parts[0] == "GEOINFORMED":
            return url_parts[0] + '_' + url_parts[3]
        else:
            return url_parts[0]

    @staticmethod
    def _get_product_id(request):
        collection = Stat._get_collection(request)
        request = Stat._clean_request(request)
        url_parts = request.split('/')
        import re
        if re.match('Sentinel2_*', collection):
            return url_parts[5]
        elif re.match('CGS_S1_SLC_L1', collection):
            return url_parts[6]
        elif re.match('CGS_S1_GRD_L1', collection):
            return url_parts[7]
        elif re.match('PROBAV_.*_V1', collection) \
                or re.match('CGS_S1_GRD_SIGMA0', collection) \
                or re.match('S1_', collection) \
                or re.match('CGS_S2_L(1C|2A)', collection):
            return url_parts[4]
        elif re.match('PROBAV_.*_V2', collection):
            if re.match("PROBAV_S", collection):
                product_id = url_parts[7].split(".")[0]
                match = re.match(r"(?P<prefix>PROBAV_S\w+_\d+K?M)_(?P<band>\w+)_(?P<version>V2\d\d)", product_id)
                if match:
                    product_id = match.group("prefix") + "_" + match.group("version")
                return product_id
            else:
                return url_parts[7]
        elif re.match('Sentinel5P_*', collection):
            return url_parts[-2]
        elif re.match('COPERNICUS-DEM-*', collection):
            return url_parts[2]
        elif re.match('COPERNICUS-EU-DEM-*', collection):
            return url_parts[2]
        elif re.match('SPOTVGT_C3_S', collection):
            return url_parts[6]
        elif re.match('SPOTVGT_C3_P', collection):
            return url_parts[7]
        elif re.match('.*WORLDCOVER.*', collection):
            return url_parts[3].split('.')[0]
        elif re.match('.*WORLDCEREAL.*', collection):
            return url_parts[6].split('.')[0]
        elif re.match('Sentinel3_S3_LST.*', collection):
            return url_parts[5].split('.')[0]
        elif re.match('Sentinel3_.*', collection):
            return url_parts[7]
        elif re.match('MAPEO_WATER_.*', collection):
            return url_parts[5]
        elif collection.startswith("GEOINFORMED_"):
            return os.path.splitext(url_parts[-1])[0]
        else:
            return '_'.join(url_parts[0:-1])


    @staticmethod
    def _get_product_date(request):
        collection = Stat._get_collection(request)
        request = Stat._clean_request(request)
        import re
        from datetime import datetime
        if re.match('Sentinel2_*', collection):
            url_parts = request.split('/')
            return datetime(int(url_parts[2]), int(url_parts[3]), int(url_parts[4]))
        elif re.match('CGS_S1_SLC_L1', collection):
            url_parts = request.split('/')
            return datetime(int(url_parts[3]), int(url_parts[4]), int(url_parts[5]))
        elif re.match('CGS_S1_GRD_L1', collection):
            url_parts = request.split('/')
            return datetime(int(url_parts[4]), int(url_parts[5]), int(url_parts[6]))
        elif re.match('S1_*', collection) \
                or re.match('CGS_S1_GRD_SIGMA0*', collection) \
                or re.match('CGS_S2_L(1C|2A)', collection):
            url_parts = request.split('/')
            return datetime(int(url_parts[1]), int(url_parts[2]), int(url_parts[3]))
        elif re.match('PROBAV_.*_V1', collection):
            url_parts = request.split('/')
            return datetime.strptime(url_parts[3], '%Y%m%d')
        elif re.match('PROBAV_.*_V2', collection):
            url_parts = request.split('/')
            i = 5 if re.match("PROBAV_S", collection) else 6
            return datetime.strptime(url_parts[i], '%Y%m%d')
        elif re.match('Sentinel5P_*', collection):
            request=request.replace("PAL__",'PAL_')     #removing the double __ from PAL_ producttypes
            product_id = Stat._get_product_id(request)
            id_parts = product_id.split('_')
            return datetime.strptime(id_parts[5], '%Y%m%d')
        elif re.match('COPERNICUS-DEM-*', collection):
            return None
        elif re.match('COPERNICUS-EU-DEM-*', collection):
            return None
        elif re.match('SPOTVGT_C3_S', collection):
            url_parts = request.split('/')
            return datetime.strptime(url_parts[5], '%Y%m%d')
        elif re.match('SPOTVGT_C3_P', collection):
            url_parts = request.split('/')
            return datetime(int(url_parts[4]), int(url_parts[5]), int(url_parts[6]))
        elif re.match('.*WORLDCOVER.*', collection):
            year = collection.split("_")[3]
            return datetime(int(year), 1, 1)
        elif re.match('.*WORLDCEREAL.*', collection):
            product_id = Stat._get_product_id(request)
            id_parts = product_id.split('_')
            return datetime.strptime(id_parts[4], '%Y-%m-%d')
        elif re.match('Sentinel3_S3_LST.*', collection):
            url_parts = request.split('/')
            return datetime.strptime(url_parts[4], '%Y%m%d')
        elif re.match('Sentinel3_.*', collection):
            url_parts = request.split('/')
            return datetime(*map(int, url_parts[4:7]))
        elif re.match('MAPEO_WATER_.*', collection):
            product_id = Stat._get_product_id(request)
            id_parts = product_id.split('_')
            return datetime.strptime(id_parts[0], '%Y%m%d')
        elif collection.startswith("GEOINFORMED_"):
            product_id = Stat._get_product_id(request)
            return datetime.strptime(product_id.split("_")[-1], "%Y-%m-%d")
        elif re.match('.*\d{8}.*', request):
            match = re.match(r".*(?P<date>\d{8}).*", request)
            if match:
                maybeDate = match.group("date")
                if int(maybeDate) > 19600101 and int(maybeDate) < 21000101:
                    return datetime.strptime(maybeDate, '%Y%m%d')
            raise Exception('Unsupported collection {} in request {}'.format(collection, request))
        else:
            raise Exception('Unsupported collection {} in request {}'.format(collection, request))


class CsvExporter(object):

    def __init__(self, csv_file):
        self.csv_file = csv_file

    def export(self, stats):
        with open(self.csv_file, 'a+') as stats_file:
            import csv
            stats_writer = csv.writer(stats_file, delimiter=',')
            for stat in stats:
                stats_writer.writerow([stat.username, stat.email, stat.timestamp, stat.collection,
                                       stat.product_date, stat.product_id, stat.product_size])


class DatabaseExporter(object):

    def __init__(self, db_host, db_name, db_username, db_password):

        self.db_host = db_host
        self.db_name = db_name
        self.db_username = db_username
        self.db_password = db_password

        try:
            import psycopg2
            self.conn = psycopg2.connect("host='{}' dbname='{}' user='{}' password='{}'"
                                    .format(self.db_host, self.db_name, self.db_username, self.db_password))
        except Exception as ex:
            LOGGER.error('Could not connect to database: {}'.format(ex))
            raise

    def get_latest_stat(self):

        from psycopg2 import sql

        cursor = None
        try:
            cursor = self.conn.cursor()
            cursor.execute(
                sql.SQL("select max(download_timestamp) from oscars_stats"))
            results = cursor.fetchone()
            return results[0]
        finally:
            if cursor is not None:
                cursor.close()

    def export(self, stats):

        from psycopg2 import sql

        cursor = None
        try:
            cursor = self.conn.cursor()
            for stat in stats:
                cursor.execute(
                    sql.SQL(
                        """INSERT INTO oscars_stats (username, email, download_timestamp, collection, product_date, product_id, product_size) 
                            VALUES (%s, %s, %s, %s, %s, %s, %s)"""),
                    [stat.username, stat.email, stat.timestamp, stat.collection, stat.product_date, stat.product_id, stat.product_size])

            self.conn.commit()
        finally:
            if cursor is not None:
                cursor.close()

    def close(self):
        if self.conn is not None:
            self.conn.close()


class StatsExporter(object):

    def __init__(self, es_hosts, es_timeout):
        self.es_hosts = es_hosts
        self.es_timeout = es_timeout

    def export_csv(self, start, end, csv_file):

        exporter = CsvExporter(csv_file)
        self._export_stats(start, end, exporter, self.es_timeout)

    def export_db(self, start, end,
                  db_host, db_name, db_username, db_password):

        exporter = DatabaseExporter(db_host, db_name, db_username, db_password)

        if start == None:
            # try to retrieve most recent download from DB
            start = exporter.get_latest_stat()

        self._export_stats(start, end, exporter, self.es_timeout)

    @staticmethod
    def _datetime2str(timestamp):
        if isinstance(timestamp, str):
            return timestamp
        return timestamp.strftime('%Y-%m-%dT%H:%M:%S.000Z')

    def _export_stats(self, start, end, exporter, timeout):

        from elasticsearch7 import Elasticsearch
        es = Elasticsearch(hosts=self.es_hosts,
                           timeout=timeout)

        es_scroll_timeout = "1m"

        LOGGER.info("Start searching download statistics")

        query = {
            "query": {
                "bool": {
                    "must": [
                        {"wildcard": {"host.name": "oscars-download*.vgt.vito.be"}},
                        {"term": {"response": 200}},
                        {"term": {"verb": "GET"}}
                    ],
                    "must_not": [
                        {"term": {"auth": "-"}},
                        {"wildcard": {"request": "*.xml"}},
                        {"wildcard": {"request": "*QUICKLOOK*"}},
                        {"wildcard": {"request": "/download/dummy/*"}},
                        {"wildcard": {"request": "/download/private/*"}},
                        {"wildcard": {"request": "/download/open/*"}},
                        {"wildcard": {"request": "/download/discovery.php*"}},
                        {"wildcard": {"clientip": "192.168.*"}},  # additional filtering on client IP-addresses
                        {"term": {"auth": "loadtest"}}  # exclude requests from the loadtest user
                    ]
                }
            }
        }

        # additional filtering on date
        if start is not None and end is not None:
            LOGGER.debug(f'Adding time filter: [{start} TO {end}]')
            from datetime import datetime
            query['query']['bool']['must'].append({"range": {"@timestamp": {"gt": StatsExporter._datetime2str(start),
                                                                            "lt": StatsExporter._datetime2str(end)}}})
        elif start is not None:
            LOGGER.debug(f'Adding time filter: [{start} TO *]')
            query['query']['bool']['must'].append({"range": {"@timestamp": {"gt": StatsExporter._datetime2str(start)}}})
        elif end is not None:
            LOGGER.debug(f'Adding time filter: [* TO {end}]')
            query['query']['bool']['must'].append({"range": {"@timestamp": {"lt": StatsExporter._datetime2str(end)}}})

        response = es.search(
            index="filebeat-index-6m*",
            body=query,
            scroll=es_scroll_timeout,
            size=100
        )

        count_hits = 0

        while response['hits']['hits']:
            es_scroll_id = response['_scroll_id']
            hits = response['hits']['hits']

            count_hits += len(hits)

            stats = []
            for hit in hits:
                LOGGER.debug(f'Converting hit {hit} to stat object')
                stat = Stat.from_es_record(hit)
                stats.append(stat)

            exporter.export(stats)

            response = es.scroll(scroll_id=es_scroll_id, scroll=es_scroll_timeout)

        LOGGER.info(f'Exported {count_hits} download statistics')
